@@ -1,11 +1,13 @@
 __all__ = ["GP"]
 
 import numpy as np
-import copy
 import matplotlib.pyplot as plt
 import scipy.optimize as optim
 import logging
 
+from scipy.linalg import cholesky, cho_solve
+from copy import copy, deepcopy
+
 from .ext import gp_c
 
 logger = logging.getLogger("gp.gp")
@@ -61,21 +63,69 @@ class GP(object):
         Initialize the GP.
 
         """
-        self._memoized = {}
-        self._x = None
-        self._y = None
-        self._s = None
-        self.n = 0
-
         #: Kernel for the gaussian process, of type
         #: :class:`~gp.kernels.base.Kernel`
         self.K = K
-        self._params = np.empty(K.params.size + 1, dtype=DTYPE)
+        self._x = None
+        self._y = None
+        self._s = None
+        self._memoized = {}
 
         self.x = x
         self.y = y
         self.s = s
 
+    def __getstate__(self):
+        state = {}
+        state['K'] = self.K
+        state['_x'] = self._x
+        state['_y'] = self._y
+        state['_s'] = self._s
+        state['_memoized'] = self._memoized
+        return state
+
+    def __setstate__(self, state):
+        self.K = state['K']
+        self._x = state['_x']
+        self._y = state['_y']
+        self._s = state['_s']
+        self._memoized = state['_memoized']
+
+    def __copy__(self):
+        state = self.__getstate__()
+        cls = type(self)
+        gp = cls.__new__(cls)
+        gp.__setstate__(state)
+        return gp
+
+    def __deepcopy__(self, memo):
+        state = deepcopy(self.__getstate__(), memo)
+        cls = type(self)
+        gp = cls.__new__(cls)
+        gp.__setstate__(state)
+        return gp
+
+    def copy(self, deep=True):
+        """
+        Create a copy of the gaussian process object.
+
+        Parameters
+        ----------
+        deep : bool (default=True)
+            Whether to return a deep or shallow copy
+
+        Returns
+        -------
+        gp : :class:`~gp.gp.GP`
+            New gaussian process object
+
+        """
+        if deep:
+            gp = deepcopy(self)
+        else:
+            gp = copy(self)
+        return gp
+
     @property
     def x(self):
         r"""
@@ -94,8 +144,8 @@ class GP(object):
     def x(self, val):
         if np.any(val != self._x):
             self._memoized = {}
-            self._x = val.astype(DTYPE)
-            self.n, = self.x.shape
+            self._x = np.array(val, copy=True, dtype=DTYPE)
+            self._x.flags.writeable = False
         else: # pragma: no cover
             pass
 
@@ -117,8 +167,9 @@ class GP(object):
     def y(self, val):
         if np.any(val != self._y):
             self._memoized = {}
-            self._y = val.astype(DTYPE)
-            if self.y.shape != (self.n,):
+            self._y = np.array(val, copy=True, dtype=DTYPE)
+            self._y.flags.writeable = False
+            if self.y.shape != self.x.shape:
                 raise ValueError("invalid shape for y: %s" % str(self.y.shape))
         else: # pragma: no cover
             pass
@@ -157,9 +208,10 @@ class GP(object):
            observation noise parameter, :math:`s`, in that order.
 
         """
-        self._params[:-1] = self.K.params
-        self._params[-1] = self._s
-        return self._params
+        _params = np.empty(self.K.params.size + 1)
+        _params[:-1] = self.K.params
+        _params[-1] = self._s
+        return _params
 
     @params.setter
     def params(self, val):
@@ -170,19 +222,16 @@ class GP(object):
         else: # pragma: no cover
             pass
 
-    def copy(self):
-        """
-        Create a copy of the gaussian process object.
+    def get_param(self, name):
+        return getattr(self.K, name)
 
-        Returns
-        -------
-        gp : :class:`~gp.gp.GP`
-            New gaussian process object
-
-        """
-        new_gp = GP(self.K.copy(), self.x, self.y, s=self.s)
-        new_gp._memoized = copy.deepcopy(self._memoized)
-        return new_gp
+    def set_param(self, name, val):
+        p = getattr(self.K, name)
+        if p != val:
+            self._memoized = {}
+            self.K.set_param(name, val)
+        else: # pragma: no cover
+            pass
 
     @memoprop
     def Kxx(self):
@@ -236,46 +285,33 @@ class GP(object):
         :math:`\mathbf{K}_{xx} = \mathbf{L}_{xx}\mathbf{L}_{xx}^\top`.
 
         """
-        return np.linalg.cholesky(self.Kxx)
-
-    @memoprop
-    def inv_Lxx(self):
-        r"""
-        Inverse cholesky decomposition of the kernel covariance matrix.
-
-        Returns
-        -------
-        inv_Lxx : numpy.ndarray
-            :math:`n\times n` matrix
-
-        Notes
-        -----
-        The value is :math:`\mathbf{L}_{xx}^{-1}`, such that:
-
-        .. math:: \mathbf{K}_{xx} = \mathbf{L}_{xx}\mathbf{L}_{xx}^\top
-
-        """
-        return np.linalg.inv(self.Lxx)
+        return cholesky(self.Kxx, lower=True, overwrite_a=False, check_finite=True)
 
     @memoprop
     def inv_Kxx(self):
         r"""
         Inverse kernel covariance matrix, :math:`\mathbf{K}_{xx}^{-1}`.
 
+        Note that this inverse is provided mostly just for
+        reference. If you actually need to use it, use the Cholesky
+        decomposition (`self.Lxx`) instead.
+
         Returns
         -------
         inv_Kxx : numpy.ndarray
             :math:`n\times n` matrix
 
         """
-        Li = self.inv_Lxx
-        return np.dot(Li.T, Li)
+        inv_Lxx = np.linalg.inv(self.Lxx)
+        return np.dot(inv_Lxx.T, inv_Lxx)
 
     @memoprop
     def inv_Kxx_y(self):
         r"""
         Dot product of the inverse kernel covariance matrix and of
         observation vector.
+        
+        This uses scipy's cholesky solver to compute the solution.
 
         Returns
         -------
@@ -287,7 +323,10 @@ class GP(object):
         This is defined as :math:`\mathbf{K}_{xx}^{-1}\mathbf{y}`.
 
         """
-        return np.dot(self.inv_Kxx, self._y)
+        inv_Kxx_y = cho_solve(
+            (self.Lxx, True), self._y, 
+            overwrite_b=False, check_finite=True)
+        return inv_Kxx_y
 
     @memoprop
     def log_lh(self):
@@ -651,5 +690,5 @@ class GP(object):
 
         ax.fill_between(X, lower, upper, color=color, alpha=0.3)
         ax.plot(X, mean, lw=2, color=color)
-        ax.plot(x, y, 'o', ms=7, color=markercolor)
+        ax.plot(x, y, 'o', ms=5, color=markercolor)
         ax.set_xlim(*xlim)
